/*  gqcam.c - The frontend to Gqcam, a GTK based QuickPict clone.
    Copyright (C) 1999 Cory Lueninghoener (cluenin1@bigred.unl.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

 */

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <pthread.h>
#include <gtk/gtk.h>
#include <linux/types.h> 
#include <linux/videodev.h>
#include "gqcam.h"
#include "frontend.h"

void init_cam(struct Camera *camera)
{
  camera->greyscale = 0;
  camera->pic = NULL;
  camera->draw = 0;
  camera->pixmap = NULL;
  camera->frozen = 0;
  camera->update_camera = 0;
  camera->saving = 0;
  camera->savetype = PPM;
}

void set_cam_info(struct Camera *camera)
{
  if (ioctl (camera->dev, VIDIOCSPICT, &camera->vid_pic) == -1) {
    perror ("ioctl (VIDIOCSPICT)");
  }
  if (ioctl (camera->dev, VIDIOCSWIN, &camera->vid_win) == -1) {
    perror ("ioctl (VIDIOCSPICT)");
  }
}

void get_cam_info(struct Camera *camera)
{
  int i;
  struct video_clip vid_clips[32];

  ioctl(camera->dev, VIDIOCGCAP, &camera->vid_caps);
  ioctl(camera->dev, VIDIOCGWIN, &camera->vid_win);
  ioctl(camera->dev, VIDIOCGPICT, &camera->vid_pic);

  for (i = 0; i < 32; i++) {
    vid_clips[i].x      = 0;
    vid_clips[i].y      = 0;
    vid_clips[i].width  = 0;
    vid_clips[i].height = 0;
  }
  camera->vid_win.clips = vid_clips;
  camera->vid_win.clipcount = 0;

  if (camera->vid_caps.type & VID_TYPE_MONOCHROME) {
    camera->greyscale = 1;
    if(camera->pic != NULL)
      free(camera->pic);
    camera->pic = malloc(camera->vid_caps.maxwidth*camera->vid_caps.maxheight);
  }
  else {
    camera->greyscale = 0;
    if (camera->pic != NULL)
      free(camera->pic);
    camera->pic = malloc(camera->vid_caps.maxwidth*camera->vid_caps.maxheight*3);
  }
}


void print_cam_info(struct Camera *camera)
{
  printf("Name: %s\n", camera->vid_caps.name);
  printf("Type: %i\n", camera->vid_caps.type);
  if (camera->vid_caps.type & VID_TYPE_CAPTURE) {
    printf("\tCan capture\n");
  }
  if (camera->vid_caps.type & VID_TYPE_TUNER) {
    printf("\tCan tune\n");
  }
  if (camera->vid_caps.type & VID_TYPE_TELETEXT) {
    printf("\tDoes teletext\n");
  }
  if (camera->vid_caps.type & VID_TYPE_OVERLAY) {
    printf("\tOverlay onto frame buffer\n");
  }
  if (camera->vid_caps.type & VID_TYPE_CHROMAKEY) {
    printf("\tOverlay by chromakey\n");
  }
  if (camera->vid_caps.type & VID_TYPE_CLIPPING) {
    printf("\tCan clip\n");
  }
  if (camera->vid_caps.type & VID_TYPE_FRAMERAM) {
    printf("\tUses the frame buffer memory\n");
  }
  if (camera->vid_caps.type & VID_TYPE_SCALES) {
    printf("\tScalable\n");
  }
  if (camera->vid_caps.type & VID_TYPE_MONOCHROME) {
    printf("\tMonochrome only\n");
  }
  if (camera->vid_caps.type & VID_TYPE_SUBCAPTURE) {
    printf("\tCan capture subareas of the image\n");
  }
  
  printf("Channels: %i\n", camera->vid_caps.channels);
  printf("Audios: %i\n", camera->vid_caps.audios);
  printf("Maxwidth: %i\n", camera->vid_caps.maxwidth);
  printf("Maxheight: %i\n", camera->vid_caps.maxheight);
  printf("Minwidth: %i\n", camera->vid_caps.minwidth);
  printf("Minheight: %i\n", camera->vid_caps.minheight);

  printf("---------\n");
  
  printf("X: %i\n", camera->vid_win.x);
  printf("Y: %i\n", camera->vid_win.y);
  printf("Width: %i\n", camera->vid_win.width);
  printf("Height: %i\n", camera->vid_win.height);
  printf("Chromakey: %i\n", camera->vid_win.chromakey);
  printf("Flags: %i\n", camera->vid_win.flags);
/*
  printf("Clips: %i\n", camera.vid_win.clips);
  printf("Clipcound: %i\n", camera.vid_win.clipcount);
*/

  printf("---------\n");
  
  printf("Brightness:\t%i (%i)\n", camera->vid_pic.brightness, camera->vid_pic.brightness/256);
  printf("Hue:\t\t%i (%i)\n", camera->vid_pic.hue, camera->vid_pic.hue/256);
  printf("Color:\t\t%i (%i)\n", camera->vid_pic.colour, camera->vid_pic.colour/256);
  printf("Contrast:\t%i (%i)\n", camera->vid_pic.contrast, camera->vid_pic.contrast/256);
  printf("Whiteness:\t%i (%i)\n", camera->vid_pic.whiteness, camera->vid_pic.whiteness/256);
  printf("Depth:\t\t%i\n", camera->vid_pic.depth);
  printf("Palette:\t%i\n", camera->vid_pic.palette);

  return;
}

void open_cam(struct Camera *camera, char *devname) 
{
  camera->dev = open(devname, O_RDWR);
  if (camera->dev < 0) {
    perror("/dev/video");
    exit(1);
  }
}

void close_cam(struct Camera *camera)
{
  close(camera->dev);
}

void display(struct Camera *camera) 
{
  GdkDrawable *drawable;
  int xpos=0, ypos=0;
  GdkRectangle update_rec;
  
  update_rec.x = 0;
  update_rec.y = 0;
  update_rec.width = 320;
  update_rec.height = 240;

  switch (camera->vid_win.width) {
  case 320:   xpos = 0;   ypos = 0;   break;
  case 160:   xpos = 80;  ypos = 60;  break;
  case 80:    xpos = 120; ypos = 90;  break;
  }

  //  printf("Displaying...\n");

  gdk_threads_enter();

  drawable = camera->drawing_area->window;
  
  if ((camera->draw)) {
    //    gdk_draw_rgb_image (drawable, camera->drawing_area->style->white_gc, xpos, ypos, camera->vid_win.width, camera->vid_win.height, GDK_RGB_DITHER_NORMAL, camera->pic, camera->vid_win.width*3);
    
    pthread_mutex_lock(&camera->pic_mutex);
    /*
    printf("waiting...");
    sleep(5);
    printf("done\n");
    */
    camera->draw = 0;
    
    if (camera->greyscale)
      gdk_draw_gray_image (camera->pixmap, camera->drawing_area->style->white_gc, xpos, ypos, camera->vid_win.width, camera->vid_win.height, GDK_RGB_DITHER_NORMAL, camera->pic, camera->vid_win.width);
    else
      gdk_draw_rgb_image (camera->pixmap, camera->drawing_area->style->white_gc, xpos, ypos, camera->vid_win.width, camera->vid_win.height, GDK_RGB_DITHER_NORMAL, camera->pic, camera->vid_win.width*3);
    
    gtk_widget_draw (camera->drawing_area, &update_rec);
    
    pthread_mutex_unlock(&camera->pic_mutex);
    
    gdk_threads_leave();
    
    //    printf("Displayed\n");
  }
  
}


void grab_image(struct Camera *camera)
{
  int len;
  
  GdkRectangle update_rec;
  GdkEventExpose *event;

  update_rec.x = 0;
  update_rec.y = 0;
  update_rec.width = 320;
  update_rec.height = 240;

  while (1) {
    if((!camera->draw) && (!camera->frozen) && (!camera->saving)){
      //    printf("Grabbing...(%i)", camera->savetype);
      
      pthread_mutex_lock(&camera->pic_mutex);

      if (camera->update_camera){
	set_cam_info(camera);
	//clearimg(camera);
	camera->update_camera = 0;
      }
      
      len = read (camera->dev, camera->pic, camera->vid_caps.maxwidth * camera->vid_caps.maxheight*3);
      //     printf("Grabbed...\n");
      if (len <= 0)
	printf("Error reading image...\n");
      
      //    pthread_mutex_lock(&camera->pic_mutex);
      camera->draw = 1;
      pthread_mutex_unlock(&camera->pic_mutex);
      
      //    gtk_signal_emit(GTK_OBJECT(camera->drawing_area), gtk_signal_lookup("expose_event",GTK_TYPE_DRAWING_AREA), camera->drawing_area, event, camera);
      
    }
  }
}

void delete_event(GtkWidget *widget, GdkEvent *event, gpointer data)
{
  gtk_main_quit ();
}


int main(int argc, char *argv[])
{
  struct Camera camera;
  pthread_t grab_thread;
  guint timeoutid;
  int i;
  char devname[256]="/dev/video";

  init_cam(&camera);

  gtk_init (&argc, &argv);
  gdk_rgb_init();

  for(i=1;i<argc;i++){
    if(strcmp(argv[i], "-h")==0 || strcmp(argv[i], "-help")==0 || strcmp(argv[i], "--help")==0){
      fprintf(stderr, "gqcam - GTK QuickCam Control\n");
      fprintf(stderr, "Copyright (C) 1999  Cory Lueninghoener <cluenin1@bigred.unl.edu>\n");
      fprintf(stderr, "This software comes with ABSOLUTELY NO WARRANTY\n");
      fprintf(stderr, "This software is free software, and you are welcome to redistribute it\n");
      fprintf(stderr, "under certain conditions\n");
      fprintf(stderr, "See the README file for a more complete notice.\n");
      fprintf(stderr, "usage:\n\n   %s [options]\n\noptions:\n\n",argv[0]);
      fprintf(stderr, "   -h | -help | --help    display this help screen\n");
      fprintf(stderr, "   -d <device>            grab frames from <device>\n");
      exit(0);
    }
    
    if(strcmp(argv[i], "-d")==0){  
      if(i<argc-1){
	i++;  
	sprintf(devname, "%s", argv[i]);
      }
    }  
  }
  
    

  open_cam(&camera, devname);
  get_cam_info(&camera);
  //  set_cam_info(&camera);
  //print_cam_info(&camera);

  create_frontend(&camera);

  pthread_create(&grab_thread, NULL, (void *)&grab_image, (void*)&camera);

   //  timeoutid = gtk_idle_add(display, (gpointer)&camera);
  timeoutid = gtk_timeout_add(500, (GtkFunction)display, (gpointer)&camera);
  
  gtk_main();

  close_cam(&camera);

  return 0;
}
